// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#pragma once
#include "HLSLMath.h"
#include "GBuffer.h"
#include <vector>
#include <thread>
#include <atomic>
#include <algorithm>
#include <chrono>

/*! Turns a G-buffer into a shaded image by invoking a given shading functor 
	per pixel.
	\param rvOutImage On termination, this will be a vector with one entry per 
			pixel of the G-buffer providing the color that resulted from 
			shading. It uses scan-line order. Left-to-right then top to 
			bottom.
	\param rGBuffer The G-buffer for which shading should be performed.
	\param nThread The number of used threads. Pass e.g. 
			std::thread::hardware_concurrency() or 1 to have all work done by 
			the current thread.
	\param rShadingFunction A shading function compatible with the calling 
			signature 
			Float3(GBuffer::Pixel pixel,Float3 position,Float3 normal,Float3 color,bool isEmissive).
	\param pOutShadingTime If this is not nullptr, it will be set to the time 
			it took to do the shading, excluding memory allocation.*/
template<class ShadingFunctionFunctor>
void shadeGBuffer(std::vector<Float3>& rvOutImage,const GBuffer& rGBuffer,std::size_t nThread,std::chrono::nanoseconds* pOutShadingTime,const ShadingFunctionFunctor& rShadingFunction){
	GBuffer::Size resolution=rGBuffer.getResolution();
	std::size_t nPixel=resolution[0]*resolution[1];
	rvOutImage.resize(nPixel);
	// Record the starting time
	auto startTime=std::chrono::system_clock::now();
	if(nThread<=1){
		for(std::size_t iPixel=0;iPixel!=nPixel;++iPixel){
			// Do the shading
			auto colorInformation=rGBuffer.getPixelColor(iPixel);
			rvOutImage[iPixel]=rShadingFunction(rGBuffer.unravelIndex(iPixel),rGBuffer.getPixelPosition(iPixel),rGBuffer.getPixelNormal(iPixel),colorInformation.first,colorInformation.second);
		}
	}
	else{
		// Threads write one batch of pixels at a time before grabbing the next 
		// batch
		std::size_t nPixelPerBatch=4096;
		std::size_t nBatch=(nPixel+nPixelPerBatch-1)/nPixelPerBatch;
		std::atomic<std::size_t> iNextBatch(0ul);
		// Spawn the threads
		std::thread::hardware_concurrency();
		std::vector<std::unique_ptr<std::thread>> threadPool;
		threadPool.reserve(nThread);
		for(std::size_t i=0;i!=nThread;++i){
			threadPool.emplace_back(new std::thread(
				[&iNextBatch,nPixelPerBatch,nBatch,nPixel,&rGBuffer,&rvOutImage,&rShadingFunction](){
					// Grab the next batch
					std::size_t iBatch;
					while((iBatch=(iNextBatch++))<nBatch){
						// Iterate over the pixels
						std::size_t iPixelEnd=std::min(nPixel,(iBatch+1)*nPixelPerBatch);
						for(std::size_t iPixel=iBatch*nPixelPerBatch;iPixel!=iPixelEnd;++iPixel){
							// Do the shading
							auto colorInformation=rGBuffer.getPixelColor(iPixel);
							rvOutImage[iPixel]=rShadingFunction(rGBuffer.unravelIndex(iPixel),rGBuffer.getPixelPosition(iPixel),rGBuffer.getPixelNormal(iPixel),colorInformation.first,colorInformation.second);
						}
					}
				}
			));
		}
		// Wait for all threads to finish
		for(auto& rpThread:threadPool){
			rpThread->join();
		}
	}
	auto endTime=std::chrono::system_clock::now();
	if(pOutShadingTime){
		(*pOutShadingTime)=endTime-startTime;
	}
}
