// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#pragma once
#include <istream>
#include <array>
#include <vector>
#include <cstddef>
#include <cstdint>
#include "HLSLMath.h"

/*! A precomputed G-buffer. The raw data is made available through getters 
	providing the world-space attributes needed for shading explicitly.*/
class GBuffer{
public:
	/*! A width and height in pixels.*/
	typedef std::array<std::size_t,2> Size;
	/*! The location of a pixel as x and y-coordinate in pixels.*/
	typedef std::array<std::size_t,2> Pixel;

	/*! Consumes binary data from the given input stream to load a previously 
		computed G-buffer among with meta-data about the camera.*/
	GBuffer(std::istream& rGBufferFile){
		rGBufferFile.read(reinterpret_cast<char*>(&mResolution),sizeof(mResolution));
		rGBufferFile.read(reinterpret_cast<char*>(&mppProjectionToWorldSpace[0][0]),sizeof(float)*16);
		rGBufferFile.read(reinterpret_cast<char*>(&mCameraPosition),sizeof(float)*3);
		mvDepth.resize(mResolution[0]*mResolution[1]);
		rGBufferFile.read(reinterpret_cast<char*>(mvDepth.data()),sizeof(float)*mvDepth.size());
		for(std::vector<std::uint32_t>* pvBuffer:{&mvColor,&mvNormal}){
			pvBuffer->resize(mResolution[0]*mResolution[1]);
			rGBufferFile.read(reinterpret_cast<char*>(pvBuffer->data()),sizeof(std::uint32_t)*pvBuffer->size());
		}
		mpViewportTransform[0]= 2.0f/mResolution[0];
		mpViewportTransform[1]=-2.0f/mResolution[1];
		mpViewportTransform[2]=0.5f*mpViewportTransform[0]-1.0f;
		mpViewportTransform[3]=0.5f*mpViewportTransform[1]+1.0f;
	}

	/*! Retrieves the width and height of this g-buffer in pixels.*/
	inline Size getResolution() const{
		return mResolution;
	}
	/*! Iterates over all pixels to count how many of them are diffuse (i.e. 
		they require non-trivial shading work).*/
	std::size_t getDiffusePixelCount() const{
		std::size_t nDiffusePixel=0;
		for(std::uint32_t value:mvColor){
			nDiffusePixel+=((value&0xC0000000)!=0)?0:1;
		}
		return nDiffusePixel;
	}
	/*! \return The world space position of the camera, inferred from the 
				projection to world space transform.*/
	inline Float3 getWorldSpaceCameraPosition(){
		return mCameraPosition;
	}

	/*! \return The world-space location of the fragment at the given pixel.
		\sa ravelIndex() */
	inline Float3 getPixelPosition(std::size_t iPixel) const{
		float depth=mvDepth[iPixel];
		Pixel pixel=unravelIndex(iPixel);
		float pProjectionSpacePosition[4]={
			pixel[0]*mpViewportTransform[0]+mpViewportTransform[2],
			pixel[1]*mpViewportTransform[1]+mpViewportTransform[3],
			depth,1.0f
		};
		float pWorldSpacePosition[4]={0.0f,0.0f,0.0f,0.0f};
		for(std::size_t i=0;i!=4;++i){
			for(std::size_t j=0;j!=4;++j){
				pWorldSpacePosition[i]+=pProjectionSpacePosition[j]*mppProjectionToWorldSpace[j][i];
			}
		}
		return Float3(pWorldSpacePosition[0]/pWorldSpacePosition[3],pWorldSpacePosition[1]/pWorldSpacePosition[3],pWorldSpacePosition[2]/pWorldSpacePosition[3]);
	}
	/*! \return The world-space normal of the fragment at the given pixel.
		\sa ravelIndex() */
	inline Float3 getPixelNormal(std::size_t iPixel) const{
		std::uint32_t value=mvNormal[iPixel];
		return Float3(-1.0f,-1.0f,-1.0f)+2.0f*Float3(
			static_cast<float>(value&0x3FF)*(1.0f/1023.0f),
			static_cast<float>(value&(0x3FF<<10))*(1.0f/(1023.0f*1024.0f)),
			static_cast<float>(value&(0x3FF<<20))*(1.0f/(1023.0f*1024.0f*1024.0f)));
	}
	/*! \return The color associated with the fragment at the given pixel and 
		its meaning. true indicates are pure emissive fragment with given 
		radiance, false indicates a diffuse fragment with given albedo.
		\sa ravelIndex() */
	inline std::pair<Float3,bool> getPixelColor(std::size_t iPixel) const{
		std::uint32_t value=mvColor[iPixel];
		return {Float3(
			static_cast<float>(value&0x3FF)*(1.0f/1023.0f),
			static_cast<float>(value&(0x3FF<<10))*(1.0f/(1023.0f*1024.0f)),
			static_cast<float>(value&(0x3FF<<20))*(1.0f/(1023.0f*1024.0f*1024.0f))),
			(value&0xC0000000)!=0
		};
	}

	/*! Turns a pixel-location into a flat index ranging from 0 to the number 
		of pixels in this buffer minus one.*/
	inline std::size_t ravelIndex(Pixel pixel) const{
		return pixel[0]+pixel[1]*mResolution[0];
	}
	/*! Inverse of ravelIndex().*/
	inline Pixel unravelIndex(std::size_t index) const{
		return {index%mResolution[0],index/mResolution[0]};
	}

private:
	/*! Width and height in pixels.*/
	Size mResolution;
	/*! Multiplying a 2D pixel index by entry 0 and 1 and adding entry 2 and 3 
		component-wise yields the projection-space xy-coordinate at w=1.0f for 
		for locations seen in the center of this pixel.*/
	float mpViewportTransform[4];
	/*! The projection to world space matrix used for rendering this G-buffer. 
		It is to be multiplied from the right.*/
	float mppProjectionToWorldSpace[4][4];
	/*! The world space location of the camera that captured this G-buffer.*/
	Float3 mCameraPosition;
	/*! The flattened depth buffer, storing z/w as fixed point number in the 24 
		least significant bits.*/
	std::vector<float> mvDepth;
	/*! The color buffer, storing 3*10 bits for RGB and a boolean for 
		emissivity in the last two bits.*/
	std::vector<std::uint32_t> mvColor;
	/*! The normal buffer, storing 3*10 bits for x, y, z.*/
	std::vector<std::uint32_t> mvNormal;
};
