// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


__import Raytracing;
__import SamplingSphericalCaps;
__import SphereLightUtility;
__import DisneyDiffuseBRDF;
__import BlueNoiseUtility;
__import PSCMaps;

// Uncomment this to do shading with the baseline technique used for run time 
// measurements in the paper. The shading result is not meaningful because the 
// BRDF is evaluated with a nonsensical incoming direction constructed from the 
// blue noise and no density is divided out.
//#define SAMPLE_BASELINE

//! The render target is bound as UAV
RWTexture2D<float4> gRenderTarget;
//! The G-buffer for the depth
Texture2D<float> gGBufferDepth;
//! The G-buffer for the diffuse albedo
Texture2D<float4> gGBufferDiffuse;
//! The G-buffer for the world-space normal
Texture2D<float4> gGBufferNormal;

//! For shadow rays, we use an empty payload
struct ShadowRayPayload{
	bool hit;
};

shared cbuffer PerFrameCB:register(b0){
	/*! The factor applied to radiance values before they are output.*/
	float gExposure;
	/*! The linear roughness parameter for the Disney diffuse BRDF used on all 
		materials in the scene.*/
	float gDisneyDiffuseRoughness;
	/*! An array of sphere lights to use for shading. The number of such lights 
		has to be given as compile-time constant (typically a define).*/
	SphereLightData gpSphereLight[SPHERE_LIGHT_COUNT];
	/*! The world space position of the camera.*/
	float3 gWorldSpaceCameraPosition;
	/*! The transform from projection space to world space.*/
	float4x4 gProjectionToWorldSpace;
	/*! The width and height of the viewport in pixels.*/
	uint2 gViewportSize;
	/*! Multiplying a 2D pixel index by gViewportTransform.xy and adding 
		gViewportTransform.zw yields the projection-space xy-coordinate at 
		w=1.0f for locations seen in the center of this pixel.*/
	float4 gViewportTransform;
};


/*! This function sends a shadow ray towards the given light source in the 
	given normalized world space incoming direction. If an occluder is found, it 
	returns zero, otherwise it forwards to evaluateDisneyDiffuseBRDF(). It 
	assumes that the ray actually points towards the light source.*/
float3 evaluateVisibilityTimesBRDF(SphereLightData sphereLight,MinimalistShadingData shadingData,float3 incomingLightDirection){
#if TRACE_SHADOW_RAYS
	// Trace a shadow ray to the point where the ray first intersects the sphere
	RayDesc ray;
	ray.Origin=shadingData.position;
	ray.TMin=1.0e-2f;
	ray.Direction=incomingLightDirection;
	float3 sphereOffset=ray.Origin-sphereLight.posW;
	float pHalf=dot(ray.Direction,sphereOffset);
	float q=dot(sphereOffset,sphereOffset)-sphereLight.radius*sphereLight.radius;
	ray.TMax=-pHalf-sqrt(max(0.0f,mad(pHalf,pHalf,-q)));
	ShadowRayPayload payload;
	payload.hit=true;
	TraceRay(gRtScene,RAY_FLAG_ACCEPT_FIRST_HIT_AND_END_SEARCH,0xFF,0,0,0,ray,payload);
	if(payload.hit){
		return float3(0.0f,0.0f,0.0f);
	}
#endif
	return evaluateDisneyDiffuseBRDF(shadingData,incomingLightDirection);
}

/*! A helper function to feed Urena's method. Given a surface normal, the 
	vector from the shading point to the center of the sphere light and the 
	radius of the sphere light, it constructs a local frame that meets the 
	requirements of this technique and expresses the spherical cap in terms of 
	angles.*/
void getUrenaCoordinateFrame(out float3 outTangent,out float3 outBitangent,out float outAlpha,out float outBeta,float3 normal,float3 relativeSphereCenter,float radius){
	float invCenterDistance=rsqrt(dot(relativeSphereCenter,relativeSphereCenter));
	float3 normalizedCenter=invCenterDistance*relativeSphereCenter;
	float normalizedCenterElevation=dot(normal,normalizedCenter);
	outTangent=mad(-normalizedCenterElevation,normal,normalizedCenter);
	float invTangentLength=rsqrt(dot(outTangent,outTangent));
	outTangent*=invTangentLength;
	outBitangent=cross(normal,outTangent);
	outAlpha=asin(min(1.0f,max(-1.0f,radius*invCenterDistance)));
	outBeta=asin(min(1.0f,max(-1.0f,normalizedCenterElevation)));
}

/*! Evaluates shading for a spherical light source using a Monte Carlo estimate 
	with a single sample. You need to pass two uniform random numbers between 
	zero and one and some of the output of getUrenaCoordinateFrame() and 
	initializePSCMap().*/
float3 evaluateSphereLightShadingProjectedSolidAngleUrena(MinimalistShadingData shadingData,SphereLightData sphereLight,PSCMaps pscMap,float3 tangent,float3 bitangent,float2 randomNumbers){
	float invDensity=pscMap.F;
	// Sample a direction towards the spherical cap
	T localX,localY;
	rad_map(pscMap,randomNumbers.x,randomNumbers.y,localX,localY);
	float3 incomingLightDirection;
	incomingLightDirection.xy=float2(localX,localY);
	incomingLightDirection.z=sqrt(saturate(mad(-incomingLightDirection.x,incomingLightDirection.x,mad(-incomingLightDirection.y,incomingLightDirection.y,1.0f))));
	incomingLightDirection=incomingLightDirection.x*tangent+incomingLightDirection.y*bitangent+incomingLightDirection.z*shadingData.normal;
	// Trace a shadow ray and evaluate the BRDF
	float3 brdf=evaluateVisibilityTimesBRDF(sphereLight,shadingData,incomingLightDirection);
	// Compute the contribution to the Monte Carlo estimate
	return brdf*sphereLight.surfaceRadiance*invDensity;
}

/*! Evaluates shading for a spherical light source using a Monte Carlo estimate 
	with a single sample. You need to pass two uniform random numbers between 
	zero and one and the output of prepareSphericalCapSampling().*/
float3 evaluateSphereLightShadingProjectedSolidAngle(MinimalistShadingData shadingData,SphereLightData sphereLight,ProjectedSphericalCap cap,float2 randomNumbers){
	// Sample a direction towards the spherical cap
	float density;
	float3 incomingLightDirection=sampleProjectedSphericalCap(density,cap,randomNumbers);
	// Trace a shadow ray and evaluate the BRDF
	float3 brdf=evaluateVisibilityTimesBRDF(sphereLight,shadingData,incomingLightDirection);
	// Compute the contribution to the Monte Carlo estimate
	return brdf*sphereLight.surfaceRadiance/density;
}


/*! Like evaluateSphereLightShading() but takes a sample from the entire 
	spherical cap uniformly (not in proportion to projected solid angle and not 
	restricted to the upper hemisphere).*/
float3 evaluateSphereLightShadingSolidAngle(MinimalistShadingData shadingData,SphereLightData sphereLight,SphericalCap cap,float2 randomNumbers){
	// Sample a direction towards the spherical cap
	float3 incomingLightDirection=sampleSphericalCap(cap,randomNumbers);
	// Early out, if the sample is in the lower hemisphere
	float lambertIncoming=dot(shadingData.normal,incomingLightDirection);
	if(lambertIncoming<=0.0f){
		return float3(0.0f,0.0f,0.0f);
	}
	// Evaluate the BRDF
	float3 brdf=evaluateVisibilityTimesBRDF(sphereLight,shadingData,incomingLightDirection);
	// Compute the contribution to the Monte Carlo estimate
	return (lambertIncoming*cap.solidAngle)*brdf*sphereLight.surfaceRadiance;
}


/*! Reads from the G-buffer to completely fill the shading data for the 
	specified pixel.*/
void getShadingDataFromGBuffer(out MinimalistShadingData outShadingData,uint2 pixel){
	// Reconstruct the world-space position from the depth
	float depth=gGBufferDepth.Load(int3(pixel,0));
	float4 projectionSpacePosition=float4(mad((float2)pixel,gViewportTransform.xy,gViewportTransform.zw),depth,1.0f);
	float4 worldSpacePosition=mul(projectionSpacePosition,gProjectionToWorldSpace);
	worldSpacePosition/=worldSpacePosition.w;
	outShadingData.position=worldSpacePosition.xyz;
	// Read and renormalize the normal (to diminish rounding error)
	outShadingData.normal=normalize(mad(gGBufferNormal.Load(int3(pixel,0)).xyz,2.0f,-1.0f));
	// Read the diffuse albedo or the emitted radiance
	float4 colorData=gGBufferDiffuse.Load(int3(pixel,0));
	float3 black=float3(0.0f,0.0f,0.0f);
	outShadingData.isPureEmissive=(colorData.a>0.0f);
	outShadingData.color=colorData.rgb;
	// Compute redundant geometric quantities
	outShadingData.outgoingLightDirection=normalize(gWorldSpaceCameraPosition-outShadingData.position);
	outShadingData.lambertOutgoing=dot(outShadingData.normal,outShadingData.outgoingLightDirection);
	// We use a uniform roughtness
	outShadingData.linearRoughness=gDisneyDiffuseRoughness;
}

[shader("miss")]
void shadowMiss(inout ShadowRayPayload payload){
	payload.hit=false;
}
[shader("anyhit")]
void shadowAnyHit(inout ShadowRayPayload payload,BuiltInTriangleIntersectionAttributes attributes){
	payload.hit=true;
}


[shader("raygeneration")]
void main(){
	// Do early out for pixels outside the viewport
	uint3 dispatchRaysIndex=DispatchRaysIndex();
	uint2 pixel=uint2(dispatchRaysIndex.x,dispatchRaysIndex.y);
	if(pixel.x>gViewportSize.x || pixel.y>gViewportSize.y){
		return;
	}
	// Extract shading data from the G-buffer
	MinimalistShadingData shadingData;
	getShadingDataFromGBuffer(shadingData,pixel);
	// For pixels that are emissive only (light sources and background), we can 
	// skip shading
	float4 finalColor=float4(0.0f,0.0f,0.0f,1.0f);
	if(shadingData.isPureEmissive){
		finalColor.rgb=shadingData.color/gExposure;
	}
	else{
		// Shade with each sphere light. Unrolling loops with ray tracing 
		// instructions gives a huge performance benefit, probably because 
		// multiple rays can be traced simultaneously for better latency 
		// hiding, hence the [unroll] and the defines.
		float4 fullBlueNoise;
		uint iHalfBlueNoiseSample=0;
#if TRACE_SHADOW_RAYS
		[unroll]
#else
		[loop]
#endif
		for(uint i=0;i!=SPHERE_LIGHT_COUNT;++i){
			SphereLightData sphereLight=gpSphereLight[i];
			// Analyze the spherical cap and terminate early, if it is entirely below 
			// the horizon
#ifndef SAMPLE_BASELINE
	#if SAMPLE_PROJECTED_SOLID_ANGLE_URENA
			float3 tangent,bitangent;
			float alpha,beta;
			getUrenaCoordinateFrame(tangent,bitangent,alpha,beta,shadingData.normal,sphereLight.posW-shadingData.position,sphereLight.radius);
			PSCMaps pscMap=initializePSCMap(alpha,beta);
			if(pscMap.invisible){
				continue;
			}
	#elif SAMPLE_PROJECTED_SOLID_ANGLE
			ProjectedSphericalCap projectedCap;
			prepareProjectedSphericalCapSampling(projectedCap,shadingData.normal,sphereLight.posW-shadingData.position,sphereLight.radius);
			if(isProjectedSphericalCapEmpty(projectedCap)){
				continue;
			}
	#else
			SphericalCap cap;
			prepareSphericalCapSampling(cap,sphereLight.posW-shadingData.position,sphereLight.radius);
	#endif
#endif
#if TRACE_SHADOW_RAYS
			[unroll]
#else
			[loop]
#endif
			for(uint j=0;j!=SAMPLE_COUNT;++j){
				float2 blueNoise;
				if(iHalfBlueNoiseSample%2==0){
					fullBlueNoise=getBlueNoiseSample64(pixel,iHalfBlueNoiseSample/2);
					blueNoise=fullBlueNoise.xy;
				}
				else{
					blueNoise=fullBlueNoise.zw;
				}
				++iHalfBlueNoiseSample;
#ifdef SAMPLE_BASELINE
				finalColor.rgb+=sphereLight.surfaceRadiance*evaluateVisibilityTimesBRDF(sphereLight,shadingData,normalize(float3(blueNoise.x,0.01f,blueNoise.y)));
#else
	#if SAMPLE_PROJECTED_SOLID_ANGLE_URENA
				finalColor.rgb+=evaluateSphereLightShadingProjectedSolidAngleUrena(shadingData,sphereLight,pscMap,tangent,bitangent,blueNoise);
	#elif SAMPLE_PROJECTED_SOLID_ANGLE
				finalColor.rgb+=evaluateSphereLightShadingProjectedSolidAngle(shadingData,sphereLight,projectedCap,blueNoise);
	#else
				finalColor.rgb+=evaluateSphereLightShadingSolidAngle(shadingData,sphereLight,cap,blueNoise);
	#endif
#endif
			}
		}
		finalColor.rgb/=SAMPLE_COUNT;
	}
	finalColor.rgb*=gExposure;
	// Write the final color to the appropriate pixel in the UAV
	gRenderTarget[pixel]=finalColor;
}
